package plan9

import Chisel._

import java.io.PrintWriter
import java.io.File

class Carp(carpParams: CarpParameters) extends Module {
  // IO
  val io = new Bundle {
    val comm = new CommIO(carpParams.TxBufferAddressBits, carpParams.RxBufferAddressBits).flip
  }

  // Constants

  // Internal signals
  val run = Bool()

  // Modules
  val fetch = Module(new Fetch(carpParams))

  val decode = Module(new Decode(carpParams))

  val sendBufferMux = Module(new SendBufferMux(carpParams))

  val ruleWriter = Module(new RuleWriter(carpParams))

  val ruleNumbersReader = Module(new RuleNumbersReader(carpParams))

  val ruleVectorReader = Module(new RuleVectorReader(carpParams))

  val cellWriterReader = Module(new CellWriterReader(carpParams))

  val lutWriter = Module(new LUTWriter(carpParams))

  val informationSender = Module(new InformationSender(carpParams))

  val fitness = Module(new FitnessLiveCount(carpParams))

  val fitnessBuffer = Module(new Fifo(
    log2Up(carpParams.FitnessBufferSize),
    32
  ))

  val fitnessSender = Module(new FitnessSender(carpParams))

  val resetter = Module(new Resetter())

  val cellStorageMux = Module(new CellStorageMux(carpParams))

  val cellStorage = Module(new CellStorage(carpParams))

  val development = Module(new Development(carpParams))

  val cellularAutomata = Module(new CellularAutomata(carpParams))

  val readout = Module(new Readout(carpParams))

  val readoutBuffer = Module(new Fifo(
    log2Up(carpParams.ReadoutBufferSize),
    1
  ))

  val readoutSender = Module(new ReadoutSender(carpParams))

  run := (fetch.io.done
    & informationSender.io.done
    & cellWriterReader.io.done
    & development.io.done
    & ruleVectorReader.io.done
    & ruleNumbersReader.io.done
    & fitnessSender.io.done
    & readoutSender.io.done
    & cellularAutomata.io.done
  )

  fetch.io.run := run
  decode.io.run := run
  informationSender.io.run := run
  cellWriterReader.io.run := run
  cellStorageMux.io.run := run
  sendBufferMux.io.run := run
  ruleWriter.io.run := run
  ruleVectorReader.io.run := run
  ruleNumbersReader.io.run := run
  lutWriter.io.run := run
  fitnessSender.io.run := run
  cellularAutomata.io.run := run
  readoutSender.io.run := run
  cellStorage.io.run := run
  resetter.io.run := run
  development.io.run := run

  // Wiring
  // Fetch
  fetch.io.bufferData <> io.comm.rx_buffer_data
  fetch.io.bufferCount <> io.comm.rx_buffer_count
  fetch.io.bufferRead <> io.comm.rx_buffer_read

  // Decode <> Fetch
  decode.io.instruction <> fetch.io.instruction

  // Rule Writer <> Decode
  ruleWriter.io.decodeOperation := decode.io.ruleWriterOperation
  ruleWriter.io.decodeAddress := decode.io.ruleWriterAddress
  ruleWriter.io.decodeData := decode.io.ruleWriterData

  // Decode <> Rule Numbers Reader
  decode.io.ruleNumbersReaderOperation <> ruleNumbersReader.io.decodeOperation

  // Decode <> Rule Vector Reader
  decode.io.ruleVectorReaderOperation <> ruleVectorReader.io.decodeOperation
  decode.io.ruleVectorReaderCount <> ruleVectorReader.io.decodeCount

  // Decode <> Cell Writer Reader
  decode.io.cellWriterReaderOperation <> cellWriterReader.io.decodeOperation
  decode.io.cellWriterReaderAddressZ <> cellWriterReader.io.decodeAddressZ
  decode.io.cellWriterReaderAddressY <> cellWriterReader.io.decodeAddressY
  decode.io.cellWriterReaderAddressX <> cellWriterReader.io.decodeAddressX
  decode.io.cellWriterReaderState <> cellWriterReader.io.decodeState
  decode.io.cellWriterReaderStates <> cellWriterReader.io.decodeStates
  decode.io.cellWriterReaderType <> cellWriterReader.io.decodeType
  decode.io.cellWriterReaderTypes <> cellWriterReader.io.decodeTypes

  // Decode <> LUT Writer
  lutWriter.io.decodeOperation := decode.io.lutWriterOperation
  lutWriter.io.decodeAddress := decode.io.lutWriterAddress
  lutWriter.io.decodeData := decode.io.lutWriterData

  // Decode <> Information Sender
  informationSender.io.decodeOperation := decode.io.informationSenderOperation

  // Decode <> Fitness Sender
  decode.io.fitnessSenderOperation <> fitnessSender.io.decodeOperation

  // Readout Sender <> Decode
  readoutSender.io.decodeOperation := decode.io.readoutSenderOperation

  // Decode <> Resetter
  resetter.io.operation := decode.io.resetterOperation

  // Decode <> Cell Storage Mux
  decode.io.cellStorageMuxSelect <> cellStorageMux.io.sourceSelect

  // Decode <> Cell Storage
  decode.io.cellStorageOperation <> cellStorage.io.operation

  // Decode <> Send Buffer Mux
  decode.io.sendBufferMuxSelect <> sendBufferMux.io.sourceSelect

  // Decode <> Development
  decode.io.developmentOperation <> development.io.decodeOperation
  decode.io.developmentRulesActive <> development.io.decodeRulesActive

  // Decode <> Cellular Automata
  cellularAutomata.io.decodeOperation := decode.io.cellularAutomataOperation
  cellularAutomata.io.decodeStepCount := decode.io.cellularAutomataStepCount

  // Decode <> Readout
  readout.io.decodeOperation := decode.io.readoutOperation
  readout.io.weightAddr := decode.io.readoutWeightAddress
  readout.io.weightData := decode.io.readoutWeightData

  // Information Sender <> Send Buffer Mux
  sendBufferMux.io.informationSenderData := informationSender.io.sendBufferData
  sendBufferMux.io.informationSenderWrite := informationSender.io.sendBufferWrite
  informationSender.io.sendBufferCount := sendBufferMux.io.informationSenderCount

  // Cell Writer Reader <> Cell Storage Mux
  cellWriterReader.io.bufferTypesIn <> cellStorageMux.io.writerReaderTypesOut
  cellWriterReader.io.bufferTypesOut <> cellStorageMux.io.writerReaderTypesIn
  cellWriterReader.io.bufferStatesIn <> cellStorageMux.io.writerReaderStatesOut
  cellWriterReader.io.bufferStatesOut <> cellStorageMux.io.writerReaderStatesIn
  cellWriterReader.io.bufferAddressZ <> cellStorageMux.io.writerReaderAddressZ
  cellWriterReader.io.bufferAddressY <> cellStorageMux.io.writerReaderAddressY
  cellWriterReader.io.bufferTypesWrite <> cellStorageMux.io.writerReaderTypesWrite
  cellWriterReader.io.bufferStatesWrite <> cellStorageMux.io.writerReaderStatesWrite

  // Send Buffer Mux <> Cell Writer Reader
  sendBufferMux.io.cellWriterReaderData <> cellWriterReader.io.sendBufferData
  sendBufferMux.io.cellWriterReaderWrite <> cellWriterReader.io.sendBufferWrite
  sendBufferMux.io.cellWriterReaderCount <> cellWriterReader.io.sendBufferCount

  // Send Buffer Mux <> Rule Numbers Reader
  sendBufferMux.io.ruleNumbersReaderData <> ruleNumbersReader.io.sendBufferData
  sendBufferMux.io.ruleNumbersReaderCount <> ruleNumbersReader.io.sendBufferCount
  sendBufferMux.io.ruleNumbersReaderWrite <> ruleNumbersReader.io.sendBufferWrite

  // Send Buffer Mux <> Comm
  sendBufferMux.io.sendBufferData <> io.comm.tx_buffer_data
  sendBufferMux.io.sendBufferWrite <> io.comm.tx_buffer_write
  sendBufferMux.io.sendBufferCount <> io.comm.tx_buffer_count

  // Send Buffer Mux <> Rule Vector Reader
  sendBufferMux.io.ruleVectorReaderData <> ruleVectorReader.io.sendBufferData
  sendBufferMux.io.ruleVectorReaderWrite <> ruleVectorReader.io.sendBufferWrite
  sendBufferMux.io.ruleVectorReaderCount <> ruleVectorReader.io.sendBufferCount

  // Send Buffer Mux <> Fitness Sender
  sendBufferMux.io.fitnessSenderData <> fitnessSender.io.sendBufferData
  sendBufferMux.io.fitnessSenderWrite <> fitnessSender.io.sendBufferWrite
  sendBufferMux.io.fitnessSenderCount <> fitnessSender.io.sendBufferCount

  // Send Buffer Mux <> Readout Sender
  sendBufferMux.io.readoutSenderData <> readoutSender.io.sendBufferData
  sendBufferMux.io.readoutSenderWrite <> readoutSender.io.sendBufferWrite
  readoutSender.io.sendBufferCount := sendBufferMux.io.readoutSenderCount

  // FitnessBuffer <> Fitness Sender
  fitnessBuffer.io.dataRead <> fitnessSender.io.fitnessBufferRead
  fitnessBuffer.io.dataOut <> fitnessSender.io.fitnessBufferData
  fitnessBuffer.io.dataCount <> fitnessSender.io.fitnessBufferCount

  // FitnessBuffer <> Fitness
  fitnessBuffer.io.dataWrite <> fitness.io.fitnessBufferWrite
  fitnessBuffer.io.dataIn <> fitness.io.fitnessBufferData
  fitnessBuffer.io.dataCount <> fitness.io.fitnessBufferCount

  fitnessBuffer.reset := resetter.io.bufferReset

  // Fitness <> FitnessSender
  fitness.io.wordsPerRun <> fitnessSender.io.fitnessWordsPerRun

  // Fitness <> Information Sender
  informationSender.io.fitnessIdentifier := fitness.io.identifier
  informationSender.io.fitnessWordsPerRun := fitness.io.wordsPerRun
  informationSender.io.fitnessParameters := fitness.io.parameters

  // ReadoutBuffer <> Readout Sender
  readoutBuffer.io.dataRead := readoutSender.io.readoutBufferRead
  readoutSender.io.readoutBufferData := readoutBuffer.io.dataOut
  readoutSender.io.readoutBufferCount := readoutBuffer.io.dataCount

  // ReadoutBuffer <> Readout
  readoutBuffer.io.dataWrite <> readout.io.outWrite
  readoutBuffer.io.dataIn <> readout.io.output

  readoutBuffer.reset <> resetter.io.bufferReset

  // Cell Storage Mux <> Cell Storage
  cellStorageMux.io.bramAAddressZ <> cellStorage.io.aAdressZ
  cellStorageMux.io.bramAAddressY <> cellStorage.io.aAdressY
  cellStorageMux.io.bramATypesWrite <> cellStorage.io.aTypesWrite
  cellStorageMux.io.bramATypesIn <> cellStorage.io.aTypesOut
  cellStorageMux.io.bramATypesOut <> cellStorage.io.aTypesIn
  cellStorageMux.io.bramAStatesWrite <> cellStorage.io.aStatesWrite
  cellStorageMux.io.bramAStatesIn <> cellStorage.io.aStatesOut
  cellStorageMux.io.bramAStatesOut <> cellStorage.io.aStatesIn

  cellStorageMux.io.bramBAddressZ <> cellStorage.io.bAdressZ
  cellStorageMux.io.bramBAddressY <> cellStorage.io.bAdressY
  cellStorageMux.io.bramBTypesWrite <> cellStorage.io.bTypesWrite
  cellStorageMux.io.bramBTypesIn <> cellStorage.io.bTypesOut
  cellStorageMux.io.bramBTypesOut <> cellStorage.io.bTypesIn
  cellStorageMux.io.bramBStatesWrite <> cellStorage.io.bStatesWrite
  cellStorageMux.io.bramBStatesIn <> cellStorage.io.bStatesOut
  cellStorageMux.io.bramBStatesOut <> cellStorage.io.bStatesIn

  // Development <> Cell Storage Mux
  development.io.bramAAddressZ <> cellStorageMux.io.developmentAAddressZ
  development.io.bramAAddressY <> cellStorageMux.io.developmentAAddressY
  development.io.bramATypesWrite <> cellStorageMux.io.developmentATypesWrite
  development.io.bramATypesIn <> cellStorageMux.io.developmentATypesOut
  development.io.bramATypesOut <> cellStorageMux.io.developmentATypesIn
  development.io.bramAStatesWrite <> cellStorageMux.io.developmentAStatesWrite
  development.io.bramAStatesIn <> cellStorageMux.io.developmentAStatesOut
  development.io.bramAStatesOut <> cellStorageMux.io.developmentAStatesIn

  development.io.bramBAddressZ <> cellStorageMux.io.developmentBAddressZ
  development.io.bramBAddressY <> cellStorageMux.io.developmentBAddressY
  development.io.bramBTypesWrite <> cellStorageMux.io.developmentBTypesWrite
  development.io.bramBTypesIn <> cellStorageMux.io.developmentBTypesOut
  development.io.bramBTypesOut <> cellStorageMux.io.developmentBTypesIn
  development.io.bramBStatesWrite <> cellStorageMux.io.developmentBStatesWrite
  development.io.bramBStatesIn <> cellStorageMux.io.developmentBStatesOut
  development.io.bramBStatesOut <> cellStorageMux.io.developmentBStatesIn

  // Development <> Rule Writer
  development.io.ruleStorageWrite <> ruleWriter.io.ruleStorageWrite
  development.io.ruleStorageAddress <> ruleWriter.io.ruleStorageAddress
  development.io.ruleStorageData <> ruleWriter.io.ruleStorageData

  // Development <> Rule Numbers Reader
  development.io.ruleNumbersReaderAddressZ <> ruleNumbersReader.io.bufferAddressZ
  development.io.ruleNumbersReaderAddressY <> ruleNumbersReader.io.bufferAddressY
  development.io.ruleNumbersReaderData <> ruleNumbersReader.io.bufferData

  // Development <> Rule Vector Reader
  development.io.ruleVectorReaderData <> ruleVectorReader.io.vectorBufferData
  development.io.ruleVectorReaderCount <> ruleVectorReader.io.vectorBufferCount
  development.io.ruleVectorReaderRead <> ruleVectorReader.io.vectorBufferRead

  // Development <> Resetter
  development.io.ruleVectorBufferReset <> resetter.io.bufferReset

  // Cellular Automata <> Cell Storage Mux
  cellularAutomata.io.bramAddressZ <> cellStorageMux.io.cellularAutomataAddressZ
  cellularAutomata.io.bramAddressY <> cellStorageMux.io.cellularAutomataAddressY
  cellularAutomata.io.bramTypesWrite <> cellStorageMux.io.cellularAutomataTypesWrite
  cellularAutomata.io.bramTypesIn <> cellStorageMux.io.cellularAutomataTypesOut
  cellularAutomata.io.bramTypesOut <> cellStorageMux.io.cellularAutomataTypesIn
  cellularAutomata.io.bramStatesWrite <> cellStorageMux.io.cellularAutomataStatesWrite
  cellularAutomata.io.bramStatesIn <> cellStorageMux.io.cellularAutomataStatesOut
  cellularAutomata.io.bramStatesOut <> cellStorageMux.io.cellularAutomataStatesIn

  // Cellular Automata <> LUT Writer
  cellularAutomata.io.lutStorageWrite <> lutWriter.io.lutStorageWrite
  cellularAutomata.io.lutStorageAddress <> lutWriter.io.lutStorageAddress
  cellularAutomata.io.lutStorageData <> lutWriter.io.lutStorageData

  // Cellular Automata <> Fitness
  cellularAutomata.io.liveCountRead <> fitness.io.liveCountBufferRead
  cellularAutomata.io.liveCountData <> fitness.io.liveCountBufferData
  cellularAutomata.io.liveCountCount <> fitness.io.liveCountBufferCount

  // Cellular Automata <> Resetter
  cellularAutomata.io.liveCountReset <> resetter.io.bufferReset

  // Cellular Automata <> Readout
  readout.io.input := cellularAutomata.io.outputCells
  readout.io.run := cellularAutomata.io.readoutRun
  readout.reset := resetter.io.bufferReset
  cellularAutomata.io.feedback := readout.io.output.toBool()
}
