package plan9

import Chisel._

class CellularAutomata(carpParams: CarpParameters) extends Module {
  val cp = carpParams

  require(cp.CellStateBits == 1)

  val io = new Bundle {
    val bramAddressZ = Bits(OUTPUT, width=log2Up(cp.MatrixDepth))
    val bramAddressY = Bits(OUTPUT, width=log2Up(cp.MatrixHeight))
    val bramTypesWrite = Bool(OUTPUT)
    val bramTypesIn = Bits(INPUT, width=cp.MatrixWidth*cp.CellTypeBits)
    val bramTypesOut = Bits(OUTPUT, width=cp.MatrixWidth*cp.CellTypeBits)
    val bramStatesWrite = Bool(OUTPUT)
    val bramStatesIn = Bits(INPUT, width=cp.MatrixWidth*cp.CellStateBits)
    val bramStatesOut = Bits(OUTPUT, width=cp.MatrixWidth*cp.CellStateBits)

    val lutStorageWrite = Bool(INPUT)
    val lutStorageAddress = Bits(INPUT, width=cp.CellTypeBits)
    val lutStorageData = Bits(INPUT, width=Math.pow(2, (cp.NeighborhoodSize).toDouble).ceil.toInt)

    val liveCountRead = Bool(INPUT)
    val liveCountData = Bits(OUTPUT, width=cp.LiveCountBufferBits)
    val liveCountCount = Bits(OUTPUT, width=log2Up(cp.LiveCountBufferSize))
    val liveCountReset = Bool(INPUT)

    val feedback = Bool(INPUT)
    val outputCells = Bits(OUTPUT, width=cp.CAOutputCells.size)

    val readoutRun = Bool(OUTPUT)

    val decodeOperation = UInt(INPUT, width=log2Up(Types.cellularAutomataOperation.size))
    val decodeStepCount = UInt(INPUT, width=16)

    val run = Bool(INPUT)
    val done = Bool(OUTPUT)
  }

  val matrix = Module(new SBlockMatrix(cp))
  val readbackSelector = Module(new Selector(
                                  cp.MatrixWidth,
                                  cp.MatrixDepth*cp.MatrixHeight
                                ))
  val lutStorage = Module(new BRAM1toN(
                            cp.CellTypeBits,
                            Math.pow(2, (cp.NeighborhoodSize).toDouble).ceil.toInt,
                            cp.MatrixWidth
                          ))
  val liveCounter = Module(new LiveCounter(cp.MatrixHeight*cp.MatrixDepth*cp.MatrixWidth))
  val liveCountBuffer = Module(new Fifo(
                                 log2Up(cp.LiveCountBufferSize),
                                 log2Up(cp.MatrixDepth * cp.MatrixHeight * cp.MatrixWidth) + 1
                               ))


  val states = Enum(UInt(), List(
                      'idle, 'configureWarmup1, 'configureWarmup2, 'configureFirst, 'configure,
                      'readback,
                      'step
                    ))
  // Constants
  val shiftRegisterBits = (math.pow(2, (cp.NeighborhoodSize).toDouble).ceil.toInt / cp.LUTConfigurationBits).toInt

  // Locals
  val state = Reg(init=states('idle))
  val done = Reg(init=Bool(true))
  val addressZ = Reg(init=UInt(0, width=log2Up(cp.MatrixDepth)))
  val addressY = Reg(init=UInt(0, width=log2Up(cp.MatrixHeight)))
  val bramStatesSelected = Reg(init=UInt(0, width=log2Up(cp.MatrixDepth*cp.MatrixHeight)))
  val updateMatrix = Reg(init=Bool(false))
  val writeStates = Reg(init=Bool(false))
  val stepsRemaining = Reg(init=UInt(0, width=16))
  val LUTStorageShiftRegister = Reg(Bits(width=cp.MatrixWidth*(math.pow(2, (cp.NeighborhoodSize).toDouble).ceil.toInt)))
  val LUTStorageShiftAmount = Reg(UInt(width=log2Up(shiftRegisterBits)))
  val configurationStateSlv = Reg(Bits(width=cp.MatrixWidth*cp.CellStateBits))
  val configurationEnableSlv = Reg(Bits(width=cp.MatrixDepth*cp.MatrixHeight))
  val delayedUpdateMatrix = RegNext(updateMatrix)

  val configurationLUTSlv = Vec.tabulate(cp.MatrixWidth*cp.LUTConfigurationBits){i =>
    LUTStorageShiftRegister((i+1)*shiftRegisterBits - 1)
  }.toBits()

  switch(state) {
    is (states('idle)) {
      updateMatrix := Bool(false)
      when (io.run) {
        switch (io.decodeOperation) {
          is (Types.cellularAutomataOperation('configure)) {
            addressZ := UInt(0)
            addressY := UInt(0)
            writeStates := Bool(false)
            state := states('configureWarmup1)
            done := Bool(false)
          }
          is (Types.cellularAutomataOperation('readback)) {
            addressZ := UInt(0)
            addressY := UInt(0)
            bramStatesSelected := UInt(0)
            writeStates := Bool(true)
            state := states('readback)
            done := Bool(false)
          }
          is (Types.cellularAutomataOperation('step)) {
            when (io.decodeStepCount === UInt(0)) {
              done := Bool(true)
            }.otherwise {
              stepsRemaining := io.decodeStepCount
              done := Bool(false)
              writeStates := Bool(false)
              state := states('step)
            }
          }
        }
      }.otherwise {
        done := Bool(true)
        writeStates := Bool(false)
      }
    }
    is (states('configureWarmup1)) {
      // Types and states are output from cell buffer
      state := states('configureWarmup2)
    }
    is (states('configureWarmup2)) {
      // LUTs for given types are output from LUT storage
      state := states('configureFirst)
    }
    is (states('configureFirst)) {
      // Next cell buffer address
      if (cp.MatrixHeight == 1) {
        addressZ := addressZ + UInt(1)
        addressY := UInt(0)
      } else {
        when (addressY === UInt(cp.MatrixHeight-1)) {
          addressZ := addressZ + UInt(1)
          addressY := UInt(0)
        }.otherwise {
          addressY := addressY + UInt(1)
        }
      }
      // Copy LUTs to shift register and begin configuration of first row
      LUTStorageShiftRegister := lutStorage.io.readDataSLV
      LUTStorageShiftAmount := UInt(0)
      configurationStateSlv := io.bramStatesIn
      configurationEnableSlv := Fill(configurationEnableSlv.getWidth - 1, Bits(0)) ## Bits(1)
      state := states('configure)
    }
    is (states('configure)) {
      // Shift LUT register to get next configuration value
      LUTStorageShiftRegister := LUTStorageShiftRegister << 1
      LUTStorageShiftAmount := LUTStorageShiftAmount + UInt(1)
      // If these are the last configuration bits, proceed to next row
      when (LUTStorageShiftAmount === UInt(shiftRegisterBits - 1)) {
        // Next cell bram address
        if (cp.MatrixHeight == 1) {
          addressZ := addressZ + UInt(1)
          addressY := UInt(0)
        } else {
          when (addressY === UInt(cp.MatrixHeight-1)) {
            addressZ := addressZ + UInt(1)
            addressY := UInt(0)
          }.otherwise {
            addressY := addressY + UInt(1)
          }
        }
        // Copy LUTs to shift register and begin configuration of next row
        LUTStorageShiftRegister := lutStorage.io.readDataSLV
        LUTStorageShiftAmount := UInt(0)
        configurationStateSlv := io.bramStatesIn
        configurationEnableSlv := configurationEnableSlv << 1
        val configureDone = (addressZ === UInt(0) || Bool(cp.MatrixDepth == 1)) &&
             (addressY === UInt(0) || Bool(cp.MatrixHeight == 1))
        when (configureDone) {
          state := states('idle)
          done := Bool(true)
        }
      }
    }
    is (states('readback)) {
      // Next cell bram address
      addressY := addressY + UInt(1)
      when (addressY === UInt(cp.MatrixHeight - 1) || Bool(cp.MatrixHeight == 1)) {
        addressY := UInt(0)
        addressZ := addressZ + UInt(1)
        when (addressZ === UInt(cp.MatrixDepth - 1) || Bool(cp.MatrixDepth == 1)) {
          state := states('idle)
          done := Bool(true)
        }
      }
      bramStatesSelected := bramStatesSelected + UInt(1)
      when (bramStatesSelected === UInt((cp.MatrixDepth * cp.MatrixHeight) - 1)) {
        bramStatesSelected := UInt(0)
      }
      writeStates := Bool(true)
    }
    is (states('step)) {
      updateMatrix := Bool(true)
      stepsRemaining := stepsRemaining - UInt(1)
      when (stepsRemaining === UInt(1)) {
        state := states('idle)
        done := Bool(true)
      }
    }
  }

  // Wiring
  liveCounter.io.run := delayedUpdateMatrix
  io.readoutRun := delayedUpdateMatrix
  lutStorage.io.writeEnabled := io.lutStorageWrite
  lutStorage.io.writeAddress := io.lutStorageAddress
  lutStorage.io.writeData := io.lutStorageData
  lutStorage.io.readAddressSlv := io.bramTypesIn

  readbackSelector.io.entriesIn := matrix.io.statesSlv
  readbackSelector.io.selected := bramStatesSelected
  io.bramStatesOut := readbackSelector.io.entryOut

  liveCounter.io.cellStates := matrix.io.statesSlv
  liveCountBuffer.io.dataIn := liveCounter.io.liveCount
  liveCountBuffer.io.dataWrite := liveCounter.io.writeCount
  liveCountBuffer.io.dataRead := io.liveCountRead
  liveCountBuffer.reset := io.liveCountReset

  matrix.io.configurationStateSlv := configurationStateSlv
  matrix.io.configurationEnableSlv := configurationEnableSlv
  matrix.io.configurationLutSlv := configurationLUTSlv
  matrix.io.update := updateMatrix
  matrix.io.feedback := io.feedback

  io.liveCountData := liveCountBuffer.io.dataOut
  io.liveCountCount := liveCountBuffer.io.dataCount
  io.outputCells := (Vec.tabulate(carpParams.CAOutputCells.size){i => matrix.io.statesSlv(carpParams.coordToIndex(carpParams.CAOutputCells(i)))}).toBits
  io.done := done
  io.bramStatesWrite := writeStates
  io.bramAddressY := addressY
  io.bramAddressZ := addressZ
  io.bramTypesWrite := Bool(false)
  io.bramTypesOut := Bits(0)
}
