package plan9

import Chisel._

import card._

import java.io.PrintWriter
import java.io.File

class CommIO(tx_buffer_address_bits: Int, rx_buffer_address_bits: Int) extends Bundle {
   // datawords to be sent to the host are placed in the tx buffer
   val tx_buffer_data = Bits(INPUT, width=32);
   val tx_buffer_count = Bits(OUTPUT, width=tx_buffer_address_bits);
   val tx_buffer_write = Bool(INPUT);

   // datawords received from host are placed in the rx buffer
   val rx_buffer_data = Bits(OUTPUT, width=32);
   val rx_buffer_count = Bits(OUTPUT, width=rx_buffer_address_bits);
   val rx_buffer_read  = Bool(INPUT);
}

class Communication(apb_addr_size: Int, apb_data_size: Int, mem_addr_size: Int, mem_data_size: Int, tx_buffer_address_bits: Int, rx_buffer_address_bits: Int, reverse_payload_endian: Boolean) extends Module {

   require(mem_addr_size <= apb_data_size)
   require(mem_data_size >= 32)
   require(apb_addr_size >= 5)
   require(apb_data_size >= 32)

   // registers
   val REG_RD_DATA = Bits(0,3)
   val REG_RD_TX_BUFFER_COUNT = Bits(1,3)
   val REG_RD_TX_BUFFER_SPACE = Bits(2,3)
   val REG_RD_RX_BUFFER_COUNT = Bits(3,3)
   val REG_RD_RX_BUFFER_SPACE = Bits(4,3)
   val REG_RD_MAGIC = Bits(5,3)
   val REG_RD_WRITE_TO_MEM = Bits(6,3)

   val REG_WR_DATA = Bits(0,3)
   val REG_WR_READ_FROM_MEM = Bits(1,3)
   val REG_WR_START_BUF = Bits(2,3)
   val REG_WR_END_BUF = Bits(3,3)

   // constants
   val MAGIC = Bits("h01234567")

   // fsm states
   val st_idle = UInt(0)
   val st_read = UInt(1)
   val st_write = UInt(2)
   val st_flush_complete = UInt(3)
   val st_fetch = UInt(4)
   val st_wait_for_data = UInt(5)

   ///////////////////////////////////////////////////////////////////////////////
   // io bundle

   val io = new Bundle {
      val apb = new ApbBus(apb_addr_size, apb_data_size)
      val mem = new MemPortIo(mem_addr_size, mem_data_size)
      val comm = new CommIO(tx_buffer_address_bits, rx_buffer_address_bits)
   }

   io.apb.PSLVERR := Bool(false)
   io.apb.IRQ := Bool(false)

   ///////////////////////////////////////////////////////////////////////////////
   // registers

   val state = Reg(init=st_idle)
   val start_buf = Reg(init=Bits(0,mem_addr_size))
   val end_buf = Reg(init=Bits(0,mem_addr_size))

   ///////////////////////////////////////////////////////////////////////////////
   // rx and tx buffers

   val rxBuffer = Module(new Queue(Bits(width=32), 1 << rx_buffer_address_bits))
   val txBuffer = Module(new Queue(Bits(width=32), 1 << tx_buffer_address_bits))

   when(reset) {
      io.comm.tx_buffer_count := Bits(0)
      io.comm.rx_buffer_count := Bits(0)
   } .otherwise {
      io.comm.tx_buffer_count := txBuffer.io.count
      io.comm.rx_buffer_count := rxBuffer.io.count
   }

   io.comm.rx_buffer_data := rxBuffer.io.deq.bits

   rxBuffer.io.enq.valid := Bool(false)
   rxBuffer.io.enq.bits := io.apb.PWDATA(31,0)
   rxBuffer.io.deq.ready := io.comm.rx_buffer_read

   txBuffer.io.enq.valid := io.comm.tx_buffer_write
   txBuffer.io.enq.bits := io.comm.tx_buffer_data
   txBuffer.io.deq.ready := Bool(false)

   ///////////////////////////////////////////////////////////////////////////////
   // dram memory

   val buf_done = start_buf === end_buf

   val reqWords = Reg(init=Bits(0,32))
   val recWords = Reg(init=Bits(0,32))

   io.mem.req.valid := Bool(false)
   io.mem.req.bits.addr := start_buf
   io.mem.req.bits.write := Bool(true)
   io.mem.req.bits.data := txBuffer.io.deq.bits
   io.mem.resp.ready := Bool(true)

   ///////////////////////////////////////////////////////////////////////////////
   // fsm

   io.apb.PREADY := Bool(true)
   io.apb.PRDATA := Bits("hdeadbeef")

   io.mem.flush := Bool(false)

   switch(state) {

      //--------------------------------------------------------------------------

      is(st_idle) {
         when(io.apb.PSEL) {
            when(io.apb.PWRITE) {
               state := st_write
            } .otherwise {
               state := st_read
            }
         }
      }

      //--------------------------------------------------------------------------

      is(st_write) {
         when(io.apb.PENABLE) {
            state := st_idle

            switch(io.apb.PADDR(4,2)) {
               is(REG_WR_DATA) {
                  rxBuffer.io.enq.valid := Bool(true)
               }
               is(REG_WR_READ_FROM_MEM) {
                  io.apb.PREADY := Bool(false)
                  reqWords := Bits(0)
                  recWords := Bits(0)
                  state := st_fetch
               }
               is(REG_WR_START_BUF) {
                  start_buf := io.apb.PWDATA
               }
               is(REG_WR_END_BUF) {
                  end_buf := io.apb.PWDATA
               }
            }
         }
      }

      //--------------------------------------------------------------------------

      is(st_fetch) {
         io.apb.PREADY := Bool(false)

         // generate memory read request
         when(io.mem.req.ready && !buf_done) {
            io.mem.req.bits.write := Bool(false)
            io.mem.req.bits.addr := start_buf
            io.mem.req.valid := Bool(true)

            start_buf := start_buf + Bits(8)
            reqWords := reqWords + Bits(1)
         }

         // get memory response
         io.mem.resp.ready := rxBuffer.io.enq.ready
         when(io.mem.resp.valid && rxBuffer.io.enq.ready) {
            rxBuffer.io.enq.valid := Bool(true)
            rxBuffer.io.enq.bits := io.mem.resp.bits.data
            recWords := recWords + Bits(1)
         }

         // goto idle when done
         when(buf_done) {
            // wait for last response, then goto idle
            when(reqWords === recWords) {
               io.apb.PREADY := Bool(true)
               state := st_idle
            }
         }
      }

      //--------------------------------------------------------------------------

      is(st_read) {
         when(io.apb.PENABLE) {
            state := st_idle

            switch(io.apb.PADDR(4,2)) {
               is(REG_RD_DATA) {
                  io.apb.PRDATA := txBuffer.io.deq.bits

                  when(!txBuffer.io.deq.valid) {
                     io.apb.PREADY := Bool(false)
                     state := state
                  } .otherwise {
                     txBuffer.io.deq.ready := Bool(true)
                  }
               }
               is(REG_RD_TX_BUFFER_COUNT) {
                  io.apb.PRDATA := txBuffer.io.count
               }
               is(REG_RD_TX_BUFFER_SPACE) {
                  io.apb.PRDATA := UInt(1 << tx_buffer_address_bits) - UInt(txBuffer.io.count) - UInt(1)
               }
               is(REG_RD_RX_BUFFER_COUNT) {
                  io.apb.PRDATA := rxBuffer.io.count
               }
               is(REG_RD_RX_BUFFER_SPACE) {
                  io.apb.PRDATA := UInt(1 << rx_buffer_address_bits) - UInt(rxBuffer.io.count) - UInt(1)
               }
               is(REG_RD_MAGIC) {
                  io.apb.PRDATA := MAGIC
               }
               is(REG_RD_WRITE_TO_MEM) {
                  io.apb.PREADY := Bool(false)
                  state := st_wait_for_data
               }
            }
         }
      }

      //--------------------------------------------------------------------------

      is(st_wait_for_data) {
         io.apb.PREADY := Bool(false)
         io.apb.PRDATA := Bits(0)

         // receive words from txBuffer and push to memory
         when(txBuffer.io.deq.valid && io.mem.req.ready && !buf_done) {
            // write to mem
            io.mem.req.valid := Bool(true)
            txBuffer.io.deq.ready := Bool(true)
            start_buf := start_buf + Bits(8)
         }

         when(buf_done) {
            io.mem.flush := Bool(true)

            when(io.mem.flush_complete) {
               io.apb.PREADY := Bool(true)
               state := st_idle;
            } .otherwise {
               state := st_flush_complete
            }
         }
      }

      //--------------------------------------------------------------------------

      is(st_flush_complete) {
         io.apb.PRDATA := Bits(0)
         io.apb.PREADY := Bool(false)
         when(io.mem.flush_complete) {
            io.apb.PREADY := Bool(true)
            state := st_idle;
         }
      }
   }
}
