package card

import Chisel._
import plan9._

import java.io.PrintWriter
import java.io.File

///////////////////////////////////////////////////////////////////////////////

class DispIO extends Bundle {
   val inst_vld = Bool(INPUT)
   val inst = Bits(INPUT, width=5)

   val aeg_idx = Bits(INPUT, width=18)
   val aeg_rd = Bool(INPUT)
   val aeg_wr = Bool(INPUT)
   val aeg_wr_data = Bits(INPUT, width=64)

   val aeg_cnt = Bits(OUTPUT, width=18)
   val exception = Bits(OUTPUT, width=16)
   val idle = Bool(OUTPUT)
   val rtn_data_vld = Bool(OUTPUT)
   val rtn_data = Bits(OUTPUT, width=64)
   val stall = Bool(OUTPUT)
}

class CsrIO extends Bundle {
   val wr_vld = Bool(INPUT)
   val rd_vld = Bool(INPUT)
   val address = Bits(INPUT, width=16)
   val wr_data = Bits(INPUT, width=64)
   val rd_ack = Bool(OUTPUT)
   val rd_data = Bits(OUTPUT, width=64)
}

class McIO extends Bundle {
   val NUM_MC_PORTS = 1
   val RTNCTL_WIDTH = 48

   val rq_vld = Bits(OUTPUT, width=NUM_MC_PORTS*1 )
   val rq_rtnctl = Bits(OUTPUT, width=NUM_MC_PORTS*RTNCTL_WIDTH)
   val rq_data = Bits(OUTPUT, width=NUM_MC_PORTS*64)
   val rq_vadr = Bits(OUTPUT, width=NUM_MC_PORTS*48)
   val rq_size = Bits(OUTPUT, width=NUM_MC_PORTS*2 )
   val rq_cmd = Bits(OUTPUT, width=NUM_MC_PORTS*3 )
   val rq_scmd = Bits(OUTPUT, width=NUM_MC_PORTS*4 )
   val rq_stall = Bits(INPUT, width=NUM_MC_PORTS*1 )

   val rs_vld = Bits(INPUT, width=NUM_MC_PORTS*1 )
   val rs_cmd = Bits(INPUT, width=NUM_MC_PORTS*3 )
   val rs_scmd = Bits(INPUT, width=NUM_MC_PORTS*4 )
   val rs_data = Bits(INPUT, width=NUM_MC_PORTS*64)
   val rs_rtnctl = Bits(INPUT, width=NUM_MC_PORTS*RTNCTL_WIDTH)
   val rs_stall = Bits(OUTPUT, width=NUM_MC_PORTS*1 )

   val rq_flush = Bits(OUTPUT, width=NUM_MC_PORTS*1 )
   val rs_flush_cmplt = Bits(INPUT, width=NUM_MC_PORTS*1 )
}

class ApbBus(addr_size: Int, data_size: Int) extends Bundle {
   val PADDR = Bits(INPUT, width = addr_size)
   val PWRITE = Bool(INPUT)
   val PSEL = Bool(INPUT)
   val PENABLE = Bool(INPUT)
   val PWDATA = Bits(INPUT, width = data_size)
   val PRDATA = Bits(OUTPUT, width = data_size)
   val PREADY = Bool(OUTPUT)
   val PSLVERR = Bool(OUTPUT)
   val IRQ = Bool(OUTPUT)
}

class MemPortIo(addr_size: Int, data_size: Int) extends Bundle {
   val req = Decoupled(new MemBus(addr_size, data_size))
   val resp = (new DecoupledIO(new MemBus(addr_size, data_size))).flip
   val flush = Bool(OUTPUT)
   val flush_complete = Bool(INPUT)
}

class MemBus(addr_size: Int, data_size: Int) extends Bundle {
   val addr = Bits(width = addr_size)
   val write = Bool()
   val data = Bits(width = data_size)
   override def clone = { new MemBus(addr_size, data_size).asInstanceOf[this.type] }
}

///////////////////////////////////////////////////////////////////////////////

class ConveyApbBridge extends Module {

   val io = new Bundle {
      val clkhx = Bool(INPUT)
      val clk2x = Bool(INPUT)

      val disp = new DispIO
      val mc = new McIO
      val csr = new CsrIO

      val i_aeid = Bits(INPUT, width=4)
   }

   val toplevel = Module(new Toplevel(32,64,64,64))

   ////////////////////////////////////////////////////////////////////////////
   // CSR
   //
   // Currently not used for anything, just two registers available
   // for demonstrating the functionality.
   //
   // Possible usecase is to toggle the io.disp.idle state.  AEGs can
   // not be used when io.disp.idle===false, so if AEGs are to be used
   // while running a system that has access to the memory controller,
   // CSRs must be used to temporarily halt the running system and
   // thereby allowing AEG access.

   val CSR_REG_MAGIC = UInt(0)
   val CSR_REG_DATA = UInt(1)

   val CSR_MAGIC = Bits("haabbccdd", 64)

   val csrData = Reg(init=Bits(0, 64))

   io.csr.rd_ack := Bool(false)
   io.csr.rd_data := Bits(0)

   when(Reg(next=io.csr.rd_vld)) {
      // must answer the cycle after getting the read
      io.csr.rd_ack := Bool(true) 

      switch(Reg(next=io.csr.address)) {
         is(CSR_REG_MAGIC) {
            io.csr.rd_data := CSR_MAGIC
         }
         is(CSR_REG_DATA) {
            io.csr.rd_data := csrData
         }
      }
   }
   when(io.csr.wr_vld) {
      // must accept data the cycle after getting the write
      switch(io.csr.address) {
         is(CSR_REG_DATA) {
            csrData := io.csr.wr_data
         }
      }
   }

   ////////////////////////////////////////////////////////////////////////////
   // MC to MemPortIo Bridge

   val MC_READ = Bits(1)
   val MC_WRITE = Bits(2)

   io.mc.rq_vld := toplevel.io.mem.req.valid
   io.mc.rq_rtnctl := toplevel.io.mem.req.bits.addr
   io.mc.rq_vadr := toplevel.io.mem.req.bits.addr
   io.mc.rq_size := Bits(3)
   io.mc.rq_scmd := Bits(0)
   io.mc.rs_stall := !toplevel.io.mem.resp.ready

   io.mc.rq_flush := toplevel.io.mem.flush
   toplevel.io.mem.flush_complete := io.mc.rs_flush_cmplt

   io.mc.rq_data := toplevel.io.mem.req.bits.data

   io.mc.rq_cmd := MC_READ
   when(toplevel.io.mem.req.bits.write) {
      io.mc.rq_cmd := MC_WRITE
   }

   toplevel.io.mem.req.ready := !io.mc.rq_stall
   toplevel.io.mem.resp.valid := io.mc.rs_vld
   toplevel.io.mem.resp.bits.addr := io.mc.rs_rtnctl
   toplevel.io.mem.resp.bits.write := Bool(false)
   toplevel.io.mem.resp.bits.data := io.mc.rs_data

   ////////////////////////////////////////////////////////////////////////////
   // Dispatch and AEG
   //
   // Drives the APB bus

   io.disp.aeg_cnt := Bits(256)
   io.disp.exception := Bits(0)
   io.disp.exception(0) := io.disp.inst_vld && (io.disp.inst =/= Bits(0))
   io.disp.exception(1) := toplevel.io.apb.IRQ
   io.disp.idle := Bool(true)
   io.disp.rtn_data_vld := Bool(false)
   io.disp.rtn_data := toplevel.io.apb.PRDATA
   io.disp.stall := Bool(false)

   val st_idle = UInt(0)
   val st_access = UInt(1)

   val state = Reg(init=st_idle);

   val aeg_idx = Reg(init=Bits(0,8))
   val aeg_write = Reg(init=Bool(false))
   val aeg_data = Reg(init=Bits(0,64))

   toplevel.io.apb.PADDR := aeg_idx ## Bits(0,2)
   toplevel.io.apb.PWRITE := aeg_write
   toplevel.io.apb.PSEL := Bool(false)
   toplevel.io.apb.PENABLE := Bool(false)
   toplevel.io.apb.PWDATA := aeg_data

   switch(state) {
      is(st_idle) {
         when(io.disp.aeg_wr || io.disp.aeg_rd) {
            aeg_idx := io.disp.aeg_idx;
            aeg_write := io.disp.aeg_wr
            aeg_data := io.disp.aeg_wr_data
            toplevel.io.apb.PADDR := io.disp.aeg_idx ## Bits(0,2)
            toplevel.io.apb.PWRITE := io.disp.aeg_wr
            toplevel.io.apb.PWDATA := io.disp.aeg_wr_data
            toplevel.io.apb.PSEL := Bool(true)
            state := st_access
         }
      }
      is(st_access) {
         io.disp.idle := Bool(false)
         io.disp.stall := Bool(true)
         toplevel.io.apb.PSEL := Bool(true)
         toplevel.io.apb.PENABLE := Bool(true)
         when(toplevel.io.apb.PREADY) {
            io.disp.idle := Bool(true)
            io.disp.rtn_data_vld := !aeg_write
            state := st_idle;
         }
      }
   }

}

///////////////////////////////////////////////////////////////////////////////

object elaborate {
  def main(args: Array[String]): Unit = {
    chiselMain(args, () => Module(new ConveyApbBridge))
  }
}
