package plan9

import Chisel._

class InformationSender(carpParams: CarpParameters) extends Module {
  val cp = carpParams

  val io = new Bundle {
    val sendBufferCount = Bits(INPUT, width=cp.TxBufferAddressBits)
    val sendBufferData = Bits(OUTPUT, width=32)
    val sendBufferWrite = Bool(OUTPUT)

    val fitnessIdentifier = Bits(INPUT, width=8)
    val fitnessWordsPerRun = Bits(INPUT, width=8)
    val fitnessParameters = Bits(INPUT, width=16)

    val decodeOperation = UInt(INPUT, width=log2Up(Types.informationSenderOperation.size))

    val run = Bool(INPUT)
    val done = Bool(OUTPUT)
  }

  val states = Enum(UInt(), List(
                      'idle, 'sendMatrixSize, 'sendCellSizeAndCounters, 'sendRuleAmount,
                      'sendFitness, 'sendReadout, 'sendReadoutInfo, 'sendReadoutTopology
                    ))

  val state = Reg(init=states('idle))

  val data = Reg(init=Bits(0, width=32))
  val write = Reg(init=Bool(false))

  val remainingWords = Reg(init=UInt(0))

  val topologyROM = Vec.tabulate(math.ceil(cp.SNNTopology.length / 2.0).toInt){i =>
    (if (i*2 + 1 >= cp.SNNTopology.length) Fill(16, Bits(0)) else Bits(cp.SNNTopology(i*2 + 1), width=16)) ##
      Bits(cp.SNNTopology(i*2), width=16)
  }

  // As long as no state sends more than 4 words, this should be enough
  val sendBufferHasSpace = UInt(cp.SendBufferSize) - io.sendBufferCount.toUInt() >= remainingWords

  switch (state) {
    is (states('idle)) {
      when (io.run && io.decodeOperation === Types.informationSenderOperation('send)) {
        state := states('sendMatrixSize)
        remainingWords := UInt(1)
      }.otherwise {
        data := Fill(32, Bits(0))
        write := Bool(false)
      }
    }
    is (states('sendMatrixSize)) {
      when (sendBufferHasSpace) {
        data := (Bits(cp.MatrixDepth, width=8) ##
                   Bits(cp.MatrixWidth, width=8) ##
                   Bits(cp.MatrixHeight, width=8) ##
                   Fill(7, Bits(0)) ## Bool(cp.MatrixWrap).toBits())
        write := Bool(true)
        state := states('sendCellSizeAndCounters)
      }.otherwise {
        write := Bool(false)
      }
    }
    is (states('sendCellSizeAndCounters)) {
      when (sendBufferHasSpace) {
        data := (Bits(cp.JumpCounterBits, width=8) ##
                   Bits(cp.JumpCounters, width=8) ##
                   Bits(cp.CellTypeBits, width=8) ##
                   Bits(cp.CellStateBits, width=8))
        write := Bool(true)
        state := states('sendRuleAmount)
      }.otherwise {
        write := Bool(false)
      }
    }
    is (states('sendRuleAmount)) {
      when (sendBufferHasSpace) {
        data := Bits(cp.RuleAmount, width=32)
        write := Bool(true)
        state := states('sendFitness)
      }.otherwise {
        write := Bool(false)
      }
    }
    is (states('sendFitness)) {
      when (sendBufferHasSpace) {
        data := (io.fitnessParameters.toBits() ##
                   io.fitnessWordsPerRun.toBits() ##
                   io.fitnessIdentifier)
        write := Bool(true)
        state := states('sendReadoutInfo)
      }.otherwise {
        write := Bool(false)
      }
    }
    is (states('sendReadoutInfo)) {
      when (sendBufferHasSpace) {
        data := (Bits(cp.SNNTopology.length, width=16) ##
                   Bits(cp.CAOutputCells.length, width=16))
        write := Bool(true)
        state := states('sendReadoutTopology)
        remainingWords := UInt(math.ceil(cp.SNNTopology.length / 2.0).toInt)
      }.otherwise {
        write := Bool(false)
      }
    }
    is (states('sendReadoutTopology)) {
      when (sendBufferHasSpace) {
        data := topologyROM(UInt(math.ceil(cp.SNNTopology.length / 2.0).toInt) - remainingWords)
        write := Bool(true)
        remainingWords := remainingWords - UInt(1)
        when (remainingWords === UInt(1)) {
          state := states('idle)
        }
      }.otherwise {
        write := Bool(false)
      }
    }
  }

  io.sendBufferData := data
  io.sendBufferWrite := write

  io.done := state === states('idle)
}
