package plan9

import Chisel._

class SBlockMatrix(carpParams: CarpParameters) extends Module {
  val cp = carpParams
  val io = new Bundle {
    val configurationLutSlv = Bits(INPUT, width=cp.MatrixWidth * cp.LUTConfigurationBits)
    val configurationStateSlv = Bits(INPUT, width=cp.MatrixWidth)
    val configurationEnableSlv = Bits(INPUT, width=cp.MatrixDepth*cp.MatrixHeight)
    val feedback = Bool(INPUT)

    val statesSlv = Bits(OUTPUT, width=cp.MatrixDepth*cp.MatrixHeight*cp.MatrixWidth)

    val update = Bool(INPUT)
  }

  val matrix = for (z <- 0 until cp.MatrixDepth) yield {
    for (y <- 0 until cp.MatrixHeight) yield {
      for (x <- 0 until cp.MatrixWidth) yield {
        if (cp.CAFeedbackCells.exists(a => a(0) == z && a(1) == y && a(2) == x)) {
          Module(new FeedbackCell())
        } else {
          Module(new SBlock(cp.NeighborhoodSize, cp.LUTConfigurationBits))
        }
      }
    }
  }

  val states = Vec.fill(cp.MatrixDepth*cp.MatrixHeight*cp.MatrixWidth){Bool()}

  for (z <- 0 until cp.MatrixDepth) {
    for (y <- 0 until cp.MatrixHeight) {
      for (x <- 0 until cp.MatrixWidth) {
        val cell = matrix(z)(y)(x)
        cell match {
          case f: FeedbackCell => {
            states(cp.coordToIndex(z, y, x)) := f.io.state
            f.io.feedback := io.feedback
            f.io.update := io.update
          }
          case s: SBlock => {
            val neighbors = cp.NeighborhoodSize - 1
            val neighbor_states = Vec.fill(neighbors){Bool()} // z z y y x x
            // X neighbor mapping
            // x_negative_nowrap
            if (x == 0 && !cp.MatrixWrap) {
              neighbor_states(neighbors-1) := states(cp.coordToIndex(z, y, x+1))
              neighbor_states(neighbors-2) := Bool(false)
            }
            // x_negative_wrap
            if (x == 0 && cp.MatrixWrap) {
              neighbor_states(neighbors-1) := states(cp.coordToIndex(z, y, x+1))
              neighbor_states(neighbors-2) := states(cp.coordToIndex(z, y, cp.MatrixWidth-1))
            }
            // x_center
            if (x > 0 && x < cp.MatrixWidth-1) {
              neighbor_states(neighbors-1) := states(cp.coordToIndex(z, y, x+1))
              neighbor_states(neighbors-2) := states(cp.coordToIndex(z, y, x-1))
            }
            // x_positive_wrap
            if (x == cp.MatrixWidth-1 && cp.MatrixWrap) {
              neighbor_states(neighbors-1) := states(cp.coordToIndex(z, y, 0))
              neighbor_states(neighbors-2) := states(cp.coordToIndex(z, y, x-1))
            }
            // x_positive_nowrap
            if (x == cp.MatrixWidth-1 && !cp.MatrixWrap) {
              neighbor_states(neighbors-1) := Bool(false)
              neighbor_states(neighbors-2) := states(cp.coordToIndex(z, y, x-1))
            }

            // Y neighbor mapping
            // y_negative_nowrap
            if (y == 0 && !cp.MatrixWrap) {
              neighbor_states(neighbors-3) := states(cp.coordToIndex(z, y+1, x))
              neighbor_states(neighbors-4) := Bool(false)
            }
            // y_negative_wrap
            if (y == 0 && cp.MatrixWrap) {
              neighbor_states(neighbors-3) := states(cp.coordToIndex(z, y+1, x))
              neighbor_states(neighbors-4) := states(cp.coordToIndex(z, cp.MatrixHeight-1, x))
            }
            // y_center
            if (y > 0 && y < cp.MatrixWidth-1) {
              neighbor_states(neighbors-3) := states(cp.coordToIndex(z, y+1, x))
              neighbor_states(neighbors-4) := states(cp.coordToIndex(z, y-1, x))
            }
            // y_positive_wrap
            if (y == cp.MatrixWidth-1 && cp.MatrixWrap) {
              neighbor_states(neighbors-3) := states(cp.coordToIndex(z, 0, x))
              neighbor_states(neighbors-4) := states(cp.coordToIndex(z, y-1, x))
            }
            // y_positive_nowrap
            if (y == cp.MatrixWidth-1 && !cp.MatrixWrap) {
              neighbor_states(neighbors-3) := Bool(false)
              neighbor_states(neighbors-4) := states(cp.coordToIndex(z, y-1, x))
            }

            // Z neighbor mapping
            // z_negative_nowrap
            if (z == 0 && !cp.MatrixWrap && cp.MatrixDepth > 1) {
              neighbor_states(neighbors-5) := states(cp.coordToIndex(z+1, y, x))
              neighbor_states(neighbors-6) := Bool(false)
            }
            // z_negative_wrap
            if (z == 0 && cp.MatrixWrap && cp.MatrixDepth > 1) {
              neighbor_states(neighbors-5) := states(cp.coordToIndex(z+1, y, x))
              neighbor_states(neighbors-6) := states(cp.coordToIndex(cp.MatrixDepth-1, y, x))
            }
            // z_center
            if (z > 0 && y < cp.MatrixWidth-1 && cp.MatrixDepth > 1) {
              neighbor_states(neighbors-5) := states(cp.coordToIndex(z+1, y, x))
              neighbor_states(neighbors-6) := states(cp.coordToIndex(z-1, y, x))
            }
            // z_positive_wrap
            if (z == cp.MatrixWidth-1 && cp.MatrixWrap && cp.MatrixDepth > 1) {
              neighbor_states(neighbors-5) := states(cp.coordToIndex(0, y, x))
              neighbor_states(neighbors-6) := states(cp.coordToIndex(z-1, y, x))
            }
            // z_positive_nowrap
            if (z == cp.MatrixWidth-1 && !cp.MatrixWrap && cp.MatrixDepth > 1) {
              neighbor_states(neighbors-5) := Bool(false)
              neighbor_states(neighbors-6) := states(cp.coordToIndex(z-1, y, x))
            }

            s.io.update := io.update
            s.io.neighborStates := neighbor_states.toBits()
            s.io.configurationLUT := io.configurationLutSlv((x+1)*cp.LUTConfigurationBits-1, x*cp.LUTConfigurationBits)
            s.io.configurationState := io.configurationStateSlv(x)
            s.io.configurationEnable := io.configurationEnableSlv(z*cp.MatrixHeight + y)
            states(cp.coordToIndex(z, y, x)) := s.io.state
          }
        }
      }
    }
  }
  io.statesSlv := states.toBits()
}
