/*****************************************************************************
 * Communication interface
 *
 * An abstraction layer above the Convey interface.
 *
 * Asbjørn Djupdal 2016
 *****************************************************************************/

#include "communication.h"

#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdbool.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

#include <wdm_user.h>

#define REG_RD_DATA 0
#define REG_RD_TX_BUFFER_COUNT 1
#define REG_RD_TX_BUFFER_SPACE 2
#define REG_RD_RX_BUFFER_COUNT 3
#define REG_RD_RX_BUFFER_SPACE 4
#define REG_RD_MAGIC 5
#define REG_RD_WRITE_TO_MEM 6

#define REG_WR_DATA 0
#define REG_WR_READ_FROM_MEM 1
#define REG_WR_START_BUF 2
#define REG_WR_END_BUF 3

#define MAGIC 0x01234567

#define BUFFER_SIZE (1024*1024)  // size of memory shared between host and FPGA, must be divisible by 8
#define WORDS_USE_RAM 5          // use shared memory for transfers larger than this number of words

static wdm_coproc_t m_coproc = WDM_INVALID;
static uint32_t *sharedMem;

static wdm_dispatch_t ds;

static void writeReg(unsigned int regInd, uint64_t regValue) {
  uint64_t reg = regValue;

  ds.ae[0].aeg_ptr_s = &reg;
  ds.ae[0].aeg_base_s = regInd;
  ds.ae[0].aeg_cnt_s = 1;
  ds.ae[0].aeg_cnt_r = 0;

  if(wdm_aeg_write_read(m_coproc, &ds) != 0) {
    fprintf(stderr, "wdm_aeg_write_read failed in writeReg\n");
    exit(1);
  }
}

static uint64_t readReg(unsigned int regInd) {
  uint64_t ret;

  ds.ae[0].aeg_cnt_s = 0;
  ds.ae[0].aeg_ptr_r = &ret;
  ds.ae[0].aeg_cnt_r = 1;
  ds.ae[0].aeg_base_r = regInd;

  if(wdm_aeg_write_read(m_coproc, &ds) != 0) {
    fprintf(stderr, "wdm_aeg_write_read failed in readReg\n");
    exit(1);
  }

  return ret;
}

void communication_open(void) {
  memset((void *)&ds, 0, sizeof(ds));

  ///////////////////////////////////////////////////////////////////////////////
  // reserve fpga

  // Reserve and attach to the coprocessor
  m_coproc = wdm_reserve(WDM_CPID_ANY, NULL);

  if (m_coproc == WDM_INVALID) {
    fprintf(stderr, "Unable to reserve coprocessor\n");
    exit(1);
  }

  ///////////////////////////////////////////////////////////////////////////////
  // attach (load bitfile)

  printf("Attaching...\n");

  char *name = "plan9";
  if (wdm_attach(m_coproc, name)) {
    fprintf(stderr, "Unable to attach signature \"%s\"\n", name);
    exit(1);
  }

  ///////////////////////////////////////////////////////////////////////////////
  // check magic

  uint32_t magic = readReg(REG_RD_MAGIC);
  if(magic != MAGIC) {
    fprintf(stderr, "Incorrect magic number %x != %x\n", magic, MAGIC);
    exit(1);
  }

  ///////////////////////////////////////////////////////////////////////////////
  // allocate memory

  wdm_posix_memalign(m_coproc, (void**)&sharedMem, 64, BUFFER_SIZE);

  ///////////////////////////////////////////////////////////////////////////////

  printf("Wolverine initialized\n");
}

void communication_close(void) {
  wdm_free(m_coproc, sharedMem);
}

void communication_send(uint64_t *buffer, int words) {
  if(words > WORDS_USE_RAM) {
    // send using shared memory
    uint64_t *buf = buffer;
    while(words > 0) {
      int size = words;
      if(size > (BUFFER_SIZE/sizeof(uint64_t))) size = BUFFER_SIZE/sizeof(uint64_t);
      wdm_memcpy(m_coproc, sharedMem, buf, size*sizeof(uint64_t));
      writeReg(REG_WR_START_BUF, (uint64_t)sharedMem);
      writeReg(REG_WR_END_BUF, (uint64_t)sharedMem+size*8);
      writeReg(REG_WR_READ_FROM_MEM, 0);

      words -= size;
      buf += size;
    }

  } else {
    // send using data register
    for(int i = 0; i < words; i++) {
      writeReg(REG_WR_DATA, buffer[i]);
    }
  }
}

void communication_receive(uint64_t *buffer, int words) {
  if(words > WORDS_USE_RAM) {
    // receive using shared memory
    uint64_t *buf = buffer;
    while(words > 0) {
      int size = words;
      if(size > (BUFFER_SIZE/sizeof(uint64_t))) size = BUFFER_SIZE/sizeof(uint64_t);

      writeReg(REG_WR_START_BUF, (uint64_t)sharedMem);
      writeReg(REG_WR_END_BUF, (uint64_t)sharedMem+size*8);
      readReg(REG_RD_WRITE_TO_MEM);
      wdm_memcpy(m_coproc, buf, sharedMem, size*sizeof(uint64_t));

      words -= size;
      buf += size;
    }

  } else {
    // receive using data register
    for(int i = 0; i < words; i++) {
      buffer[i] = readReg(REG_RD_DATA);
    }
  }
}

uint32_t communication_tx_count(void) {
  return readReg(REG_RD_TX_BUFFER_COUNT);
}

uint32_t communication_tx_space(void) {
  return readReg(REG_RD_TX_BUFFER_SPACE);
}

uint32_t communication_rx_count(void) {
  return readReg(REG_RD_RX_BUFFER_COUNT);
}

uint32_t communication_rx_space(void) {
  return readReg(REG_RD_RX_BUFFER_SPACE);
}

